#define __AVR_ATmega328P__
/**
 * Serial LCD Display With Off-Screen Buffer
 * Kerry Wong
 * http://www.kerrywong.com
 * 2/2012
 */

#include <Arduino.h>
#include <LiquidCrystal/LiquidCrystal.h>
#include <SPI/SPI.h>
#include <EEPROM/EEPROM.h>

// <editor-fold defaultstate="collapsed" desc="Pin Definitions">
const int pinRS = 2;
const int pinEn = 3;

const int pinD4 = 5;
const int pinD5 = 6;
const int pinD6 = 7;
const int pinD7 = 8;

const int BTN_CLEAR = 0;
const int BTN_UP = 1;
const int BTN_DOWN = 2;
const int BTN_LEFT = 3;
const int BTN_RIGHT = 4;

const int pinBtnClear = 4;

const int pinBtnUp = 14;
const int pinBtnDown = 15;
const int pinBtnLeft = 16;
const int pinBtnRight = 17;

const int FRAM_CS1 = 9;
const int FRAM_CS2 = 10;
// </editor-fold>

// <editor-fold defaultstate="collapsed" desc="FRAM Commands">
const byte CMD_WREN = 0x06; //0000 0110
const byte CMD_WRDI = 0x04; //0000 0100
const byte CMD_RDSR = 0x05; //0000 0101
const byte CMD_WRSR = 0x01; //0000 0001
const byte CMD_READ = 0x03; //0000 0011
const byte CMD_WRITE = 0x02; //0000 0010
// </editor-fold>


// <editor-fold defaultstate="collapsed" "EEPROM addresses for parameter storage">
const int ADDR_CUR_ROW = 0;
const int ADDR_CUR_COL = 1;
const int ADDR_CUR_SCROLL_POS = 2;
const int ADDR_CUR_LINE_BUFFER_IDX = 3;
const int ADDR_CUR_LINE_BUFFER_BOTTOM = 4;
const int ADDR_LINE_BUFFER_FULL = 5; 
// </editor-fold>

int btnPins[] = {pinBtnClear, pinBtnUp, pinBtnDown, pinBtnLeft, pinBtnRight};
int btnStates[] = {HIGH, HIGH, HIGH, HIGH, HIGH};
int previousBtnStates[] = {HIGH, HIGH, HIGH, HIGH, HIGH};
long lastDebounceTime[5];

const int SCREEN_BUFFER_SIZE = 128; //128 x 32 = 4096 bytes
const int LINE_BUFFER_LENGTH = 128; //maximum 128 rows
const int LINE_BUFFER_SIZE = 32; //each row has a maximum of 32 bytes
char lineBuf[LINE_BUFFER_SIZE];

LiquidCrystal lcd(pinRS, pinEn, pinD4, pinD5, pinD6, pinD7);
int curRow = 0, curCol = 0;
int curScrollPos = 0, curLineBufferIdx = 0, curLineBufferBottom = 0;
bool lineBufferFull = false; //whether the line buffer is full

/**
 * Write to FRAM (assuming 2 FM25C160 are used)
 * addr: starting address
 * buf: pointer to data
 * count: data length.
 *        If this parameter is omitted, it is defaulted to one byte.
 * returns: 0 operation is successful
 *          1 address out of range
 */
int FRAMWrite(int addr, byte *buf, int count = 1) {
    int cs = 0;

    if (addr > 0x7ff) {
        addr -= 0x800;
        cs = FRAM_CS2;
    } else {
        cs = FRAM_CS1;
    }

    if (addr > 0x7ff) {
        return -1;
    }

    byte addrMSB = (addr >> 8) & 0xff;
    byte addrLSB = addr & 0xff;

    digitalWrite(cs, LOW);

    SPI.transfer(CMD_WREN);
    digitalWrite(cs, HIGH);

    digitalWrite(cs, LOW);
    SPI.transfer(CMD_WRITE);
    SPI.transfer(addrMSB);
    SPI.transfer(addrLSB);

    for (int i = 0; i < count; i++) {
        SPI.transfer(buf[i]);
    }

    digitalWrite(cs, HIGH);

    return 0;
}

/**
 * Read from FRAM (assuming 2 FM25C160 are used)
 * addr: starting address
 * buf: pointer to data
 * count: data length.
 *        If this parameter is omitted, it is defaulted to one byte.
 * returns: 0 operation is successful
 *          1 address out of range
 */
int FRAMRead(int addr, byte *buf, int count = 1) {
    int cs = 0;

    if (addr > 0x7ff) {
        addr -= 0x800;
        cs = FRAM_CS2;
    } else {
        cs = FRAM_CS1;
    }

    if (addr > 0x7ff) {
        return -1;
    }

    byte addrMSB = (addr >> 8) & 0xff;
    byte addrLSB = addr & 0xff;

    digitalWrite(cs, LOW);

    SPI.transfer(CMD_READ);
    SPI.transfer(addrMSB);
    SPI.transfer(addrLSB);

    for (int i = 0; i < count; i++) {
        buf[i] = SPI.transfer(0x00);
    }

    digitalWrite(cs, HIGH);

    return 0;
}

/**
 * Read saved settings from EEPROM and set the scroll position
 */
void readSettings() {
    byte b = EEPROM.read(ADDR_LINE_BUFFER_FULL);

    if (b != 255) {
        lineBufferFull = b;

        curRow = EEPROM.read(ADDR_CUR_ROW);
        curCol = EEPROM.read(ADDR_CUR_COL);
        curScrollPos = EEPROM.read(ADDR_CUR_SCROLL_POS);
        curLineBufferIdx = EEPROM.read(ADDR_CUR_LINE_BUFFER_IDX);
        curLineBufferBottom = EEPROM.read(ADDR_CUR_LINE_BUFFER_BOTTOM);
        
        Serial.print("CurRow = "); Serial.println(curRow);
        Serial.print("CurCol = "); Serial.println(curCol);
        Serial.print("CurScrollPos = "); Serial.println(curScrollPos);
        Serial.print("CurLineBufferIdx = "); Serial.println(curLineBufferIdx);
        Serial.print("CurLineBufferBottom = "); Serial.println(curLineBufferBottom);
        
        //write to screen the lines stored at curScrollPos
        char line[LINE_BUFFER_SIZE];
        FRAMRead(curScrollPos * LINE_BUFFER_SIZE, (byte*) line, LINE_BUFFER_SIZE);
        lcd.setCursor(0, 1);
        lcd.print(line);
        
        int pos = curScrollPos - 1;
        
        if (pos < 0) pos = pos + LINE_BUFFER_LENGTH;
        FRAMRead(pos * LINE_BUFFER_SIZE, (byte*) line, LINE_BUFFER_SIZE);
        lcd.setCursor(0, 0);
        lcd.print(line);
    }
}

/**
 * Save the current screen settings into EEPROM 
 */
void saveSettings() {
    EEPROM.write(ADDR_CUR_ROW, curRow);
    EEPROM.write(ADDR_CUR_COL, curCol);
    EEPROM.write(ADDR_CUR_SCROLL_POS, curScrollPos);
    EEPROM.write(ADDR_CUR_LINE_BUFFER_IDX, curLineBufferIdx);
    EEPROM.write(ADDR_CUR_LINE_BUFFER_BOTTOM, curLineBufferBottom);
    EEPROM.write(ADDR_LINE_BUFFER_FULL, lineBufferFull);
}

void setup() {
    lcd.begin(16, 2);
    Serial.begin(9600);
    
    for (int i = 0; i < LINE_BUFFER_SIZE; i++) lineBuf[i] = ' ';
    
    for (int i = 0; i < 5; i++) {
        pinMode(btnPins[i], INPUT);
        digitalWrite(btnPins[i], HIGH);
    }

    //SPI Setup
    pinMode(FRAM_CS1, OUTPUT);
    digitalWrite(FRAM_CS1, HIGH);
    pinMode(FRAM_CS2, OUTPUT);
    digitalWrite(FRAM_CS2, HIGH);

    SPI.begin();
    SPI.setDataMode(SPI_MODE0);
    SPI.setBitOrder(MSBFIRST);
    SPI.setClockDivider(SPI_CLOCK_DIV2);
    
    readSettings();
}

/**
 * clear a row on the LCD
 * @param rowNum: the row to be cleared (either 0 or 1)
 */
void clearRow(int rowNum) {
    lcd.setCursor(0, rowNum);
    for (int i = 0; i < 16; i++) lcd.print(" ");
    for (int i = 0; i < LINE_BUFFER_SIZE; i++) lineBuf[i] = ' ';
    lcd.setCursor(0, rowNum);
}

/**
 * used for button states de-bouncing
 */
void checkBtnStates() {
    int states[5];

    for (int i = 0; i < 5; i++) {
        states[i] = digitalRead(btnPins[i]);

        if (states[i] != previousBtnStates[i]) lastDebounceTime[i] = millis();

        if (millis() - lastDebounceTime[i] > 50) {
            btnStates[i] = states[i];
        }

        previousBtnStates[i] = states[i];
    }
}

/**
 * scroll the screen up by one row
 */
void scrollUp() {
    char line[LINE_BUFFER_SIZE];

    if (curLineBufferIdx >= curLineBufferBottom) {
        if (curScrollPos > 1) {
            FRAMRead((curScrollPos - 1) * LINE_BUFFER_SIZE, (byte*) line, LINE_BUFFER_SIZE);
            lcd.setCursor(0, 1);
            lcd.print(line);
            curScrollPos--;
            FRAMRead((curScrollPos - 1) * LINE_BUFFER_SIZE, (byte*) line, LINE_BUFFER_SIZE);
            lcd.setCursor(0, 0);
            lcd.print(line);
        }
    } else {
        if (curScrollPos >= 0) {
            int pos = curScrollPos - 1;
            if (pos < 0) pos += LINE_BUFFER_LENGTH;

            FRAMRead(pos * LINE_BUFFER_SIZE, (byte*) line, LINE_BUFFER_SIZE);
            lcd.setCursor(0, 1);
            lcd.print(line);
            curScrollPos--;
            pos = curScrollPos - 1;
            if (pos < 0) pos += LINE_BUFFER_LENGTH;
            FRAMRead(pos * LINE_BUFFER_SIZE, (byte*) line, LINE_BUFFER_SIZE);
            lcd.setCursor(0, 0);
            lcd.print(line);
        } else {
            if (curScrollPos < 0) curScrollPos += LINE_BUFFER_LENGTH;

            if (curScrollPos > curLineBufferBottom) {
                FRAMRead((curScrollPos - 1) * LINE_BUFFER_SIZE, (byte*) line, LINE_BUFFER_SIZE);
                lcd.setCursor(0, 1);
                lcd.print(line);
                curScrollPos--;
                FRAMRead((curScrollPos - 1) * LINE_BUFFER_SIZE, (byte*) line, LINE_BUFFER_SIZE);
                lcd.setCursor(0, 0);
                lcd.print(line);
            }
        }
    }
}

/**
 * scroll the screen down by one row
 */
void scrollDown() {
    char line[LINE_BUFFER_SIZE];

    if (curLineBufferIdx >= curLineBufferBottom) {
        if (curScrollPos < curLineBufferIdx) {
            FRAMRead(curScrollPos*LINE_BUFFER_SIZE, (byte*) line, LINE_BUFFER_SIZE);
            lcd.setCursor(0, 0);
            lcd.print(line);
            curScrollPos++;
            FRAMRead(curScrollPos*LINE_BUFFER_SIZE, (byte*) line, LINE_BUFFER_SIZE);
            lcd.setCursor(0, 1);
            lcd.print(line);
        }
    } else {
        if (curScrollPos < LINE_BUFFER_LENGTH) {
            FRAMRead(curScrollPos*LINE_BUFFER_SIZE, (byte*) line, LINE_BUFFER_SIZE);
            lcd.setCursor(0, 0);
            lcd.print(line);
            curScrollPos++;
            if (curScrollPos == LINE_BUFFER_LENGTH) curScrollPos = 0;

            FRAMRead(curScrollPos*LINE_BUFFER_SIZE, (byte*) line, LINE_BUFFER_SIZE);
            lcd.setCursor(0, 1);
            lcd.print(line);
        } else {
            if (curScrollPos < curLineBufferBottom) {
                FRAMRead(curScrollPos*LINE_BUFFER_SIZE, (byte*) line, LINE_BUFFER_SIZE);
                lcd.setCursor(0, 0);
                lcd.print(line);
                curScrollPos++;
                FRAMRead(curScrollPos*LINE_BUFFER_SIZE, (byte*) line, LINE_BUFFER_SIZE);
                lcd.setCursor(0, 1);
                lcd.print(line);
            }
        }
    }
}

/**
 * output some debug information
 */
void dumpDebugInfo() {
    Serial.print("row=");
    Serial.print(curRow);
    Serial.print("; scroll pos=");
    Serial.print(curScrollPos);
    Serial.print("; buf index=");
    Serial.print(curLineBufferIdx);
    Serial.print("; buf bottom=");
    Serial.println(curLineBufferBottom);
}

/**
 * clear the memory and the display
 */
void clearMemory() {
    char buf[LINE_BUFFER_SIZE];

    for (int i = 0; i < LINE_BUFFER_SIZE; i++) buf[i] = ' ';
    for (int i = 0; i < LINE_BUFFER_LENGTH; i++) {
        FRAMWrite(i*LINE_BUFFER_SIZE, (byte *) buf, LINE_BUFFER_SIZE);
    }

    curRow = 0;
    curCol = 0;
    curScrollPos = 0;
    curLineBufferIdx = 0;
    curLineBufferBottom = 0;
    
    lcd.clear();
    
    curRow = 0; 
    curCol = 0;
    curScrollPos = 0;
    curLineBufferIdx = 0;
    curLineBufferBottom = 0;
    lineBufferFull = false;
    
    saveSettings();
}

/**
 * handle button clicked event
 */
void handleButtonClicks() {
    if (btnStates[BTN_RIGHT] == LOW) {
        lcd.scrollDisplayLeft();
        btnStates[BTN_RIGHT] = HIGH;
        dumpDebugInfo();

        delay(100);
    } else if (btnStates[BTN_LEFT] == LOW) {
        lcd.scrollDisplayRight();
        btnStates[BTN_LEFT] = HIGH;
        dumpDebugInfo();
        delay(100);
    } else if (btnStates[BTN_UP] == LOW) {
        scrollUp();
        btnStates[BTN_UP] = HIGH;
        dumpDebugInfo();
        delay(100);
    } else if (btnStates[BTN_DOWN] == LOW) {
        scrollDown();
        btnStates[BTN_DOWN] = HIGH;
        dumpDebugInfo();
        delay(100);
    } else if (btnStates[BTN_CLEAR] == LOW) { //button clear has two states
        if (millis() - lastDebounceTime[BTN_CLEAR] < 1000) {
            Serial.println("Saving settings");
            delay(1000);
            saveSettings();
        } else {
            Serial.println("Clear memory");
            btnStates[BTN_CLEAR] = HIGH;
            delay(500);
            clearMemory();
        }
        dumpDebugInfo();
        delay(100);
    }
}

/*
 * shift the display up by one line
 */
void shiftUp() {
    lcd.setCursor(0, 0);
    for (int i = 0; i < LINE_BUFFER_SIZE; i++) {
        lcd.write(lineBuf[i]);
        Serial.print(lineBuf[i]);
    }
    lcd.setCursor(0, 1);
    Serial.println();
}

/**
 * main loop
 */
void loop() {
    checkBtnStates();

    handleButtonClicks();

    if (Serial.available()) {
        char c = Serial.read();
        if (c != 0x0d && c != 0x0a) {
            if (curCol == 0) {
                if (curRow > 0) {
                    shiftUp();
                }
                clearRow(curRow == 0 ? 0 : 1);
            }
            lcd.write(c);

            if (curCol < LINE_BUFFER_SIZE) {
                lineBuf[curCol] = c;
            }

            curCol++;
        } else {
            FRAMWrite(curLineBufferIdx*LINE_BUFFER_SIZE, (byte*) lineBuf, LINE_BUFFER_SIZE);

            curLineBufferIdx++;
            curScrollPos = curLineBufferIdx;

            if (curLineBufferIdx >= LINE_BUFFER_LENGTH) {
                lineBufferFull = true;
                curLineBufferIdx = curLineBufferIdx % LINE_BUFFER_LENGTH;
            }

            curRow++;

            if (curRow >= LINE_BUFFER_LENGTH) {
                curLineBufferBottom = curRow % LINE_BUFFER_LENGTH + 1;
            }

            curCol = 0;

            dumpDebugInfo();
        }
    }
}
